@@ -9,11 +9,13 @@ import glob
 import numpy as np
 import astropy.io.fits as pf
 
-from nenucal import worker, futils
+from libpipe import worker, futils
 from . import database, utils
 
 CWD = os.getcwd()
 
+H5_FLAG_BACKUP = 'flags_before_ssins.h5'
+
 
 def get_all_tasks():
     d = {}
@@ -125,9 +127,9 @@ class ImagingSubTask(AbstractSubTask):
         s = f'{self.obs_id}{eo_str}-SB{self.get_sb()}-UV{img.umin}_{img.umax}_{w_str}'
         return os.path.join(self.get_dir(), s)
 
-    def register_files(self, files, lst_idx=None):
+    def register_files(self, files, lst_idx=None, img_name='image'):
         meta = self.img_rev.get_meta(self.obs_id, lst_idx=lst_idx)
-        meta.add_fits(files, even=self.even, odd=self.odd)
+        meta.add_fits(files, even=self.even, odd=self.odd, img_name=img_name)
         meta.save()
 
     def done_callback(self):
@@ -143,7 +145,13 @@ class ImagingSubTask(AbstractSubTask):
         else:
             total_time = self.obs_meta.total_time
 
-        files = glob.glob(f'{self.get_fits_prefix()}*.fits')
+        if img.clean_niter > 0 and img.clean_residual:
+            files = glob.glob(f'{self.get_fits_prefix()}*psf.fits')
+            files.extend(glob.glob(f'{self.get_fits_prefix()}*residual.fits'))
+            img_name = 'residual'
+        else:
+            files = glob.glob(f'{self.get_fits_prefix()}*.fits')
+            img_name = 'image'
 
         for file in files:
             pf.setval(file, 'PEINTTIM', value=int(self.obs_meta.int_time))
@@ -151,11 +159,11 @@ class ImagingSubTask(AbstractSubTask):
             pf.setval(file, 'PECHWIDT', value=int(self.obs_meta.chan_width))
 
         if not img.lst_bins:
-            self.register_files(files)
+            self.register_files(files, img_name=img_name)
         else:
             for i, data_lst_idx in enumerate(self.img_rev.get_lst_bins(self.obs_id, return_time_index=True)[2]):
                 files = glob.glob(f'{self.get_fits_prefix()}*-t{i:04}-*.fits')
-                self.register_files(files, lst_idx=data_lst_idx)
+                self.register_files(files, lst_idx=data_lst_idx, img_name=img_name)
 
     def get_parameters(self):
         args = []
@@ -171,6 +179,10 @@ class ImagingSubTask(AbstractSubTask):
                 channels_out = self.obs_meta.n_channels
             else:
                 channels_out = 1
+        elif str(channels_out).startswith('every'):
+            assert self.obs_meta.n_channels
+            channels_out = self.obs_meta.n_channels // int(channels_out[5:])
+        channels_out = int(channels_out)
 
         args.append("-oversampling 4095 -kernel-size 15 -nwlayers 32 -grid-mode kb -taper-edge 100")
         args.append("-padding 2 -visibility-weighting-mode unit -reorder")
@@ -191,6 +203,16 @@ class ImagingSubTask(AbstractSubTask):
             lst_bins, (start_idx, end_idx, _), _ = self.img_rev.get_lst_bins(self.obs_id, return_time_index=True)
             args.append(f'-interval {start_idx} {end_idx} -intervals-out {len(lst_bins)}')
 
+        for key, value in img.wsclean_args.items():
+            args.append(f'-{key} {value}')
+
+        if img.clean_niter > 0:
+            args.append(f'-niter {img.clean_niter} -auto-threshold {img.clean_auto_threshold}')
+            args.append(f'-local-rms -mgain {img.clean_mgain} -stop-negative -no-update-model-required')
+
+            if channels_out > 1:
+                args.append('-join-channels')
+
         args.append(f'-name {self.get_fits_prefix()} {" ".join(self.inp_ms_files)}')
 
         return args
@@ -225,7 +247,8 @@ class ImagingTask(AbstractTask):
                 return []
 
             ms_files = self.db.get_obs_id_meta(obs_id).ms_files
-            if ms_files is None:
+
+            if not ms_files:
                 print(f'Error: Obs ID {obs_id} has no MS files registered.')
                 return []
 
@@ -458,35 +481,6 @@ class MergeMSSubTask(AbstractSubTask):
         return args
 
 
-class SSINSSubTask(AbstractSubTask):
-
-    def __init__(self, settings_obj, ms_file):
-        AbstractSubTask.__init__(self, settings_obj, 'flagtool ssins')
-        self.ms_file = ms_file
-
-    def get_log_file(self):
-        return os.path.join(self.ms_file.replace('.MS', '_ssins.log'))
-
-    def get_path(self):
-        d = os.path.join(os.path.dirname(self.ms_file), 'flagger')
-        futils.mkdir(d)
-        return d
-
-    def get_config(self):
-        f = os.path.join(self.ms_file, 'ssins_config.toml')
-        self.settings.ssins.save(f)
-        return f
-
-    def get_parameters(self):
-        args = [self.ms_file, f'--config={self.get_config()} --plot_dir={self.get_path()}']
-        if self.settings.ssins.apply_ssins:
-            args.append('--backup')
-        else:
-            args.append('--restore')
-
-        return args
-
-
 class MergeMSTask(AbstractTask):
 
     name = 'merge_ms'
@@ -508,7 +502,8 @@ class MergeMSTask(AbstractTask):
                 return []
 
             ms_files = self.db.get_obs_id_meta(obs_id).ms_files
-            if ms_files is None:
+
+            if not ms_files:
                 print(f'Error: Obs ID {obs_id} has no MS files registered.')
                 return []
 
@@ -525,40 +520,6 @@ class MergeMSTask(AbstractTask):
         return all_obs_id_out
 
 
-class SSINSTask(AbstractTask):
-
-    name = 'ssins'
-    desc = 'Apply (or unapply) SSINS flagging on MS'
-
-    def __init__(self, settings_obj):
-        AbstractTask.__init__(self, 'SSINS', settings_obj, max_concurrent=1)
-        self.db = database.Database(self.settings)
-
-    def run(self, obs_ids):
-        if not self.settings.validate('ssins'):
-            return []
-
-        for obs_id in obs_ids:
-            if not self.db.has_obs_id(obs_id):
-                print(f'Error: Obs ID {obs_id} does not exist.')
-                return []
-
-            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
-            if ms_files is None:
-                print(f'Error: Obs ID {obs_id} has no MS files registered.')
-                return []
-
-            if not (len(ms_files) == 1 and len(ms_files[0]) == 1):
-                print(f'Error: Obs ID {obs_id} has invalid ms files.')
-                return []
-
-            self.add_sub_task(SSINSSubTask(self.settings, ms_files[0][0]))
-
-        self.execute()
-
-        return obs_ids
-
-
 class MakePSSubTask(AbstractSubTask):
 
     def __init__(self, settings_obj, obs_id, lst_idx=None):
@@ -856,3 +817,193 @@ class VisToSphTask(AbstractTask):
         self.execute()
 
         return new_obs_ids
+
+
+class ApplyFlagSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file, h5_file):
+        AbstractSubTask.__init__(self, settings_obj, 'flagtool restore')
+        self.ms_file = ms_file
+        self.h5_file = h5_file
+
+    def get_parameters(self):
+        return [self.ms_file, self.h5_file]
+
+
+class ApplyFlagTask(AbstractTask):
+
+    name = 'apply_flag'
+    desc = 'Apply flag h5 file'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'ApplyFlag', settings_obj)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                h5_file = os.path.join(ms_file, self.settings.apply_flag.filename)
+
+                if not os.path.exists(h5_file):
+                    print(f'Error: Flag file {h5_file} does not exist.')
+                    return []
+
+                self.add_sub_task(ApplyFlagSubTask(self.settings, ms_file, h5_file))
+
+        self.execute()
+
+        return obs_ids
+
+
+class RestoreFlagTask(AbstractTask):
+
+    name = 'restore_flag'
+    desc = 'Restore flag h5 file'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'ApplyFlag', settings_obj)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                h5_file = os.path.join(ms_file, H5_FLAG_BACKUP)
+
+                if not os.path.exists(h5_file):
+                    print(f'No backup to restore from for {ms_file}.')
+                    continue
+
+                self.add_sub_task(ApplyFlagSubTask(self.settings, ms_file, h5_file))
+
+        self.execute()
+
+        return obs_ids
+
+
+class DelayFlaggerSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file):
+        AbstractSubTask.__init__(self, settings_obj, 'flagtool delay_flag')
+        self.ms_file = ms_file
+
+    def get_log_file(self):
+        return os.path.join(self.ms_file.replace('.MS', '_delay_flagger.log'))
+
+    def get_path(self):
+        d = os.path.join(os.path.dirname(self.ms_file), 'flagger')
+        futils.mkdir(d)
+        return d
+
+    def get_parameters(self):
+        s = self.settings.delay_flagger
+        return [self.ms_file, f'--data_col={self.settings.image.data_col}', f'--umin={s.umax}', f'--umin={s.umax}',
+                f'--n_time_avg={s.n_time_avg}', f'--n_times={s.n_times}', f'--n_sigma_i={s.n_sigma_i}',
+                f'--n_sigma_v={s.n_sigma_v}', '--backup', f'--backup_file={H5_FLAG_BACKUP}']
+
+
+class DelayFlagger(AbstractTask):
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'DelayFlagger', settings_obj)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        if not self.settings.validate('delay_flagger'):
+            return []
+
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                self.add_sub_task(DelayFlaggerSubTask(self.settings, ms_file))
+
+        self.execute()
+
+        return obs_ids
+
+
+class SSINSSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file):
+        AbstractSubTask.__init__(self, settings_obj, 'flagtool ssins')
+        self.ms_file = ms_file
+
+    def get_log_file(self):
+        return os.path.join(self.ms_file.replace('.MS', '_ssins.log'))
+
+    def get_path(self):
+        d = os.path.join(os.path.dirname(self.ms_file), 'flagger')
+        futils.mkdir(d)
+        return d
+
+    def get_config(self):
+        f = os.path.join(self.ms_file, 'ssins_config.toml')
+        self.settings.ssins.save(f)
+        return f
+
+    def get_parameters(self):
+        return [self.ms_file, f'--data_col={self.settings.image.data_col}', f'--config={self.get_config()}',
+                f'--plot_dir={self.get_path()}', '--backup', f'--backup_file={H5_FLAG_BACKUP}']
+
+
+class SSINSTask(AbstractTask):
+
+    name = 'ssins'
+    desc = 'Apply SSINS flagging on MS'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'SSINS', settings_obj, max_concurrent=1)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        if not self.settings.validate('ssins'):
+            return []
+
+        if not self.settings.ssins.apply_ssins:
+            return obs_ids
+
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                self.add_sub_task(SSINSSubTask(self.settings, ms_file))
+
+        self.execute()
+
+        return obs_ids
