@@ -2,6 +2,7 @@
 #
 # Author: F. Mertens
 
+import itertools
 import astropy.time as at
 
 
@@ -27,3 +28,10 @@ def is_in_lst_bin(lst, lst_s, lst_e):
     if lst_e > lst_s:
         return (lst >= lst_s) and (lst < lst_e)
     return (lst >= lst_s) or (lst < lst_e)
+
+
+def pairwise(iterable):
+    '''s -> (s0,s1), (s1,s2), (s2, s3), ...'''
+    a, b = itertools.tee(iterable)
+    next(b, None)
+    return list(zip(a, b))
